///****************************   LICENSE   *****************************
/// Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/
/// Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at

///    http://www.apache.org/licenses/LICENSE-2.0

/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///**********************************************************************/

import 'package:analyzer/dart/constant/value.dart';
import 'package:sqfentity_base/sqfentity_base.dart';

class SqfEntityModelBuilder extends SqfEntityModelBase {
  SqfEntityModelBuilder(this.model, this.instanceName);
  final DartObject model;
  final String instanceName;
//type 'String' is not a subtype of type 'SqfEntitySequenceBase' in type cast

  SqfEntityModelBase toModel() {
    final myModel = MyDbModel()
      ..modelName = getStringValue(model, 'modelName') ?? instanceName
      ..databaseName = getStringValue(model, 'databaseName')
      ..sequences = toSequenceList(getListValue(model, 'sequences'))
      ..databaseTables = toTableList(getListValue(model, 'databaseTables'))
      ..bundledDatabasePath = getStringValue(model, 'bundledDatabasePath');

    //myModel = myModel ?? MyDbModel();
    return myModel;
  }

  SqfEntityTableBase toTable(DartObject obj) {
    return SqfEntityTableBase()
      ..tableName = getStringValue(obj, 'tableName') // getListValue(obj, 'fields').toString() 
      ..primaryKeyName = getStringValue(obj, 'primaryKeyName')
      ..primaryKeyType = getTypeValue(obj, 'primaryKeyType') as PrimaryKeyType
      ..useSoftDeleting = getBoolValue(obj, 'useSoftDeleting')
      ..fields = toFields(getListValue(obj, 'fields'))
      ..defaultJsonUrl = getStringValue(obj, 'defaultJsonUrl')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntityTableBase> toTableList(List<DartObject> objTables) {
    print('SQFENTITY_BUILDER.DART: recognizing Tables');
    final sqfEntityTableList = List<SqfEntityTableBase>();
    for (var obj in objTables) {
      sqfEntityTableList.add(toTable(obj));
    }
    return sqfEntityTableList;
  }

  SqfEntitySequenceBase toSequence(DartObject obj) {
    return SqfEntitySequenceBase()
      ..sequenceName = getStringValue(obj, 'sequenceName')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntitySequenceBase> toSequenceList(List<DartObject> objSequences) {
    print('SQFENTITY_BUILDER.DART: recognizing Sequences');
    final sqfEntitySQEList = List<SqfEntitySequenceBase>();
    for (var obj in objSequences) {
      try {
        sqfEntitySQEList.add(toSequence(obj));
      } catch (e) {
        print('SQFENTITY_BUILDER.DART: recognizing Sequences ERROR. tried obj=>${obj.toString()}\nError Message:${e.toString()}');
      }
      
    }
    return sqfEntitySQEList;
  }

  List<SqfEntityFieldType> toFields(List<DartObject> objFields) {
    final sqfEntityFieldList = List<SqfEntityFieldType>();
    for (var obj in objFields) {
      if (obj.toString().startsWith('SqfEntityFieldRelationship')) {
        sqfEntityFieldList.add(SqfEntityFieldRelationshipBase(
            toTable(obj.getField('parentTable')),
            getTypeValue(obj, 'deleteRule') as DeleteRule)
          ..defaultValue = getStringValue(obj, 'defaultValue')
        );
      } else {
        sqfEntityFieldList.add(SqfEntityFieldBase(
            getStringValue(obj, 'fieldName'),
            getTypeValue(obj, 'dbType') as DbType)
          ..defaultValue = getStringValue(obj, 'defaultValue')
          ..sequencedBy = obj
                  .getField('sequencedBy')
                  .toString()
                  .contains('SqfEntitySequenceAnnotation')
              ? toSequence(obj.getField('sequencedBy'))
              : null);
      }
    }
    return sqfEntityFieldList
        ;
  }

  String getStringValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toStringValue() : null;
  bool getBoolValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toBoolValue() : null;
  List<DartObject> getListValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toListValue() : null;
  dynamic getTypeValue(DartObject obj, String name) =>
      ifExist(obj, name) ? convertType(obj.getField(name)) : null;

  bool ifExist(DartObject obj, String name) => obj.getField(name) != null;
}

dynamic convertType(dynamic T) {
  final type = T.type;
  var types = Map<String, dynamic>();
  if (type.toString() == 'DbType') {
    types = dbTypes();
  } else if (type.toString() == 'PrimaryKeyType') {
    types = primaryKeyTypes();
  } else if (type.toString() == 'DeleteRule') {
    types = deleteRuleTypes();
  }
  for (var typ in types.entries) {
    if (T.toString().contains(typ.key)) {
      return typ.value;
    }
  }
  return type.toString();
}

Map<String, dynamic> dbTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DbType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}

Map<String, dynamic> primaryKeyTypes() {
  final types = Map<String, dynamic>();
  for (var typ in PrimaryKeyType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}

Map<String, dynamic> deleteRuleTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DeleteRule.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] =
        typ; //.add(typ.toString().substring(typ.toString().indexOf('.') + 1));
  }
  return types;
}

class MyDbModel extends SqfEntityModelBase {}
